/* 
 * keyshades
 * a tool for disabling the treo keyboard backlight
 * copyright (c) 2005, daniel grobe sachs <dgsachs@nekito.net>
 *
 * This is free software. See included LICENSE file for details.
 */

#include <PalmOS.h>
#include <palmOne_68K.h>
#include <68K/System/HsExt.h>
#include <Common/System/HsExtCommon.h>

#include "resource.h"

#define STARTUP_CODE_SIZE	152		// size of prc-tools crt0.c

#define sysNotifyKeyReplacePriority -96

#define FTR_PRESSTIME 0
#define FTR_STATE 1
		// 0 => no keydown posted
		// 1 => posted keydown
		// 2 => already changed lights
		// 3 => did launch
#define DefaultLit false	// Default to lights out
        //hsLightCircumstanceKeylightOff          = 4,    /* turn keylight off */

#define FTR_PTR	   2

static void SetLights(Boolean state);
UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags);

// Lights tools
#define ToggleLights() SetLights(!LightsAreOn())
#define LightsAreOn() (!(HsCurrentLightCircumstance() & hsLightCircumstanceKeylightOffFlagBit))

static void SetLights(Boolean state);

// Endian swapper
#define EndianSwap16(n) (((((UInt16) n) << 8) & 0xFF00) | \
		         ((((UInt16) n) >> 8) & 0x00FF))

static void RegisterForNotifications();
static void DeregisterNotifications();
static void HandleControlEvent(EventType *eventP);

// Notification handler
static Err HandleNotification(SysNotifyParamType *np)
{
	UInt32 time;
	UInt32 state;
	struct EventType newevent;
	EventType *eventP = (EventType*)np->notifyDetailsP;

	// Check if we're deleting the application
	if( np->notifyType == sysNotifyDeleteProtectedEvent )
	{
		if( ((SysNotifyDBInfoType *)np->notifyDetailsP)->creator == CREATOR )
		{
			DeregisterNotifications();
			np->handled = true;
		}

		return 0;
	}

	// Is this a queue event?
	if( np->notifyType != sysNotifyEventDequeuedEvent )
		return 0;

	// Is this a control event? (Special case for brightness bugfix)
	if( (eventP->eType == EndianSwap16(ctlSelectEvent)) || 
	    (eventP->eType == EndianSwap16(penUpEvent)) || 
	    (eventP->eType == EndianSwap16(keyDownEvent)) )
	{
		HandleControlEvent(eventP);
	}


	// Is this a keyboard event?
	if( (eventP->eType != EndianSwap16(keyUpEvent)) && (eventP->eType != EndianSwap16(keyDownEvent)) )
		return 0;

	// Is it related to the side key?
	if( eventP->data.keyUp.chr != EndianSwap16(hsChrSide) )
		return 0;

	switch( eventP->data.keyDown.modifiers )
	{
		case EndianSwap16(0x148):	// Normal release, I hope
			//SndPlaySystemSound(sndError);
			FtrGet(CREATOR,FTR_STATE,&state);

			switch( state ) 
			{	
				case 0: // do nothing
					break;

				case 1: // Need to re-send the side key keyDown and keyUp

					// Requeue the keyUp event
					MemSet(&newevent,sizeof(EventType),0);
					newevent.eType = keyUpEvent;
					newevent.data.keyUp.chr = EndianSwap16(eventP->data.keyUp.chr);
					newevent.data.keyUp.keyCode = EndianSwap16(eventP->data.keyUp.keyCode);
					newevent.data.keyUp.modifiers = 0x008; // Normal release (I hope)
					EvtAddEventToQueue(&newevent);

					// Change this event to a keyDown (it won't get reprocessed)
					eventP->eType = EndianSwap16(keyDownEvent);
					eventP->data.keyDown.modifiers = EndianSwap16(0x808);
					break;

				case 2:
				case 3:
					// we did other stuff, so eat the keyUp
					eventP->eType = EndianSwap16(nilEvent);
	
					// reset state machine
					FtrSet(CREATOR,FTR_PRESSTIME,0);
					FtrSet(CREATOR,FTR_STATE,0);
					break;

			}
			break;
				
		case EndianSwap16(0x808): // standard unmodified press
			// Save the key-press time
			FtrSet(CREATOR,FTR_PRESSTIME,TimGetTicks());
			FtrSet(CREATOR,FTR_STATE,1);
			eventP->eType = EndianSwap16(nilEvent);

			break;
				
		case EndianSwap16(0x848): // held
		case EndianSwap16(0x8C8):
			FtrGet(CREATOR,FTR_PRESSTIME,&time);
			FtrGet(CREATOR,FTR_STATE,&state);
			eventP->eType = EndianSwap16(nilEvent);

			switch( state ) 
			{
				case 1:
					// 0.5 sec to keyboard toggle
					if( TimGetTicks() > (time + SysTicksPerSecond()/2) )
					{
						ToggleLights();
						FtrSet(CREATOR,FTR_STATE,2);
					}
					break;

				case 2:
					// 1 sec to launch
					if( TimGetTicks() > (time + SysTicksPerSecond()) )
					{
						UInt32 appid;
						UInt16 cardid;
						UInt32 dbid;
						UInt32 len = sizeof(appid);
						DmSearchStateType sstate;

						// set the state
						FtrSet(CREATOR,FTR_STATE,3);
								
						// put the keyboard lights back
						ToggleLights();
							
						// launch the requested application
						HsPrefGet(hsPrefSideHardCharAppCreator, &appid, &len);
						DmGetNextDatabaseByTypeCreator(true,&sstate,'appl',appid,true,&cardid,&dbid);
						if( dbid )
							SysUIAppSwitch(cardid, dbid, sysAppLaunchCmdNormalLaunch, NULL);
					}
					break;

				default:
					break;	
					
			}
			break;

		default:	// if we get something unrecognized, stop messing with crap
			// reset state machine
			FtrSet(CREATOR,FTR_PRESSTIME,0);
			FtrSet(CREATOR,FTR_STATE,0);
			break;

	}

	return 0;
}

static void HandleControlEvent(EventType *eventP)
{
	char *title = NULL;
	FormType *frm = NULL;

	frm = FrmGetActiveForm();
	if( frm ) 
		title = FrmGetTitle(frm);

	if( title ) 
		if( !StrCompare(title, "Adjust Brightness") )
		{
			ToggleLights();
			ToggleLights();
		}
}

static void SetLights(Boolean state)
{

#if 0
	UInt16 nightmode;

	// Query night mode
	HsLightMode(false, &nightmode);
#endif

	// turn on if requested
	while( state && !LightsAreOn())
		HsLightCircumstance(false,hsLightCircumstanceKeylightOff);

	// turn off if requested
	if( !state && LightsAreOn() )
		HsLightCircumstance(true,hsLightCircumstanceKeylightOff);

#if 0
	// Restore night mode
	HsLightMode(true, &nightmode);
#endif

}

static void RegisterForNotifications()
{
	MemHandle hndl;
	void *memptr;
	void *ftrptr;
	void *hn_ptr;

	UInt16 cardNo;
	LocalID dbID;

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	hndl = DmGet1Resource('code',1);
	memptr = MemHandleLock(hndl);

	FtrPtrNew(CREATOR, FTR_PTR, MemHandleSize(hndl),&ftrptr);
	DmWrite(ftrptr, 0, memptr, MemHandleSize(hndl));

	MemHandleUnlock(hndl);
	DmDatabaseProtect(cardNo, dbID, true);

	hn_ptr = ((void *)ftrptr)+STARTUP_CODE_SIZE;

	SysNotifyRegister(cardNo, dbID, sysNotifyEventDequeuedEvent, hn_ptr, 
			sysNotifyKeyReplacePriority, NULL);

 	SysNotifyRegister(cardNo, dbID, sysNotifyDeleteProtectedEvent, NULL,
			sysNotifyNormalPriority, NULL);
}

static void DeregisterNotifications()
{	
	Int32 val;
	void *ftrptr;

	UInt16 cardNo;
	LocalID dbID;

	val = 0; FtrGet(CREATOR, FTR_PTR, &val); ftrptr = (void *)val;

	// Check if we're registered
	if( ftrptr == NULL )
		return; 

	if( 0 != SysCurAppDatabase(&cardNo, &dbID) )
		return;

	SysNotifyUnregister(cardNo, dbID, sysNotifyEventDequeuedEvent,
			sysNotifyKeyReplacePriority);

 	SysNotifyUnregister(cardNo, dbID, sysNotifyDeleteProtectedEvent,
			sysNotifyNormalPriority);

	// release feature memory
	FtrPtrFree(CREATOR, FTR_PTR);
	DmDatabaseProtect(cardNo, dbID, false);
}


//// PilotMain /////////////////////////////////////////////////////////
// It all starts here...

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
	Err err = 0;
	EventType event;

	switch( launchCode ) 
	{
		case sysAppLaunchCmdSystemReset :
			FtrSet(CREATOR,FTR_PRESSTIME,0);
			FtrSet(CREATOR,FTR_STATE,0);
			SetLights(DefaultLit);

			RegisterForNotifications();
			break;

		case sysAppLaunchCmdNotify : 
			HandleNotification((SysNotifyParamType *)cmdPBP);
			break;

		case sysAppLaunchCmdNormalLaunch :
#if 0
			FtrSet(CREATOR,FTR_PRESSTIME,0);
			FtrSet(CREATOR,FTR_STATE,0);
			SetLights(DefaultLit);

			RegisterForNotifications();
#endif

			// Well, uh, we don't have a UI, so we just quit
			MemSet( &event, sizeof(event), 0);

			event.eType = appStopEvent;
			EvtAddEventToQueue(&event);
			break;

		default:
			break;
	}


	return err;
}

